import unittest

from zope.testing import doctest
from zope import component
from zope.component import testing
from zope.annotation import interfaces as annotation_interfaces
from zope.annotation import attribute
from hurry.workflow import interfaces, workflow

class WorkflowVersions(workflow.WorkflowVersions):
    """Simplistic implementation that keeps track of versions.

    A real implementation would use something like the catalog.
    """
    def __init__(self):
        self.versions = []
        
    def addVersion(self, obj):
        self.versions.append(obj)
        
    def getVersions(self, state, id):
        result = []
        for version in self.versions:
            state_adapter = interfaces.IWorkflowState(version)
            if state_adapter.getId() == id and state_adapter.getState() == state:
                result.append(version)
        return result

    def getVersionsWithAutomaticTransitions(self):
        result = []
        for version in self.versions:
            if interfaces.IWorkflowInfo(version).hasAutomaticTransitions():
                result.append(version)
        return result
    
    def hasVersion(self, state, id):
        return bool(self.getVersions(state, id))

    def hasVersionId(self, id):
        for version in self.versions:
            state_adapter = interfaces.IWorkflowState(version)
            if state_adapter.getId() == id:
                return True
        return False
    
    def clear(self):
        self.versions = []
        
def workflowSetUp(doctest):
    testing.setUp(doctest)
    component.provideAdapter(
        workflow.WorkflowState,
        (annotation_interfaces.IAnnotatable,),
        interfaces.IWorkflowState)
    component.provideAdapter(
        workflow.WorkflowInfo,
        (annotation_interfaces.IAnnotatable,),
        interfaces.IWorkflowInfo)
    component.provideAdapter(
        attribute.AttributeAnnotations,
        (annotation_interfaces.IAttributeAnnotatable,),
        annotation_interfaces.IAnnotations)
    component.provideUtility(
        WorkflowVersions(),
        interfaces.IWorkflowVersions)
    
def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite(
            'workflow.txt',
            setUp=workflowSetUp, tearDown=testing.tearDown,
            ),
        ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')




import random, sys

from zope.interface import implements
from zope.event import notify
from zope.security.checker import CheckerPublic
from zope.security.interfaces import NoInteraction, Unauthorized
from zope.security.management import getInteraction
from zope import component

from zope.annotation.interfaces import IAnnotations
from zope.lifecycleevent import ObjectModifiedEvent
from zope.component.interfaces import ObjectEvent

from hurry.workflow import interfaces
from hurry.workflow.interfaces import MANUAL, AUTOMATIC, SYSTEM
from hurry.workflow.interfaces import\
     IWorkflow, IWorkflowState, IWorkflowInfo, IWorkflowVersions
from hurry.workflow.interfaces import\
     InvalidTransitionError, ConditionFailedError

def NullCondition(wf, context):
    return True

def NullAction(wf, context):
    pass

# XXX this is needed to make the tests pass in the absence of
# interactions..
def nullCheckPermission(permission, principal_id):
    return True

class Transition(object):

    def __init__(self, transition_id, title, source, destination,
                 condition=NullCondition,
                 action=NullAction,
                 trigger=MANUAL,
                 permission=CheckerPublic,
                 order=0,
                 **user_data):
        self.transition_id = transition_id
        self.title = title
        self.source = source
        self.destination = destination
        self.condition = condition
        self.action = action
        self.trigger = trigger
        self.permission = permission
        self.order = order
        self.user_data = user_data

    def __cmp__(self, other):
        return cmp(self.order, other.order)

# in the past this subclassed from zope.container.Contained and
# persistent.Persistent.
# to reduce dependencies these base classes have been removed.
# You can choose to create a subclass in your own code that
# mixes these in if you need persistent workflow
class Workflow(object):
    implements(IWorkflow)

    def __init__(self, transitions):
        self.refresh(transitions)

    def _register(self, transition):
        transitions = self._sources.setdefault(transition.source, {})
        transitions[transition.transition_id] = transition
        self._id_transitions[transition.transition_id] = transition

    def refresh(self, transitions):
        self._sources = {}
        self._id_transitions = {}
        for transition in transitions:
            self._register(transition)
        self._p_changed = True

    def getTransitions(self, source):
        try:
            return self._sources[source].values()
        except KeyError:
            return []

    def getTransition(self, source, transition_id):
        transition = self._id_transitions[transition_id]
        if transition.source != source:
            raise InvalidTransitionError(source)
        return transition

    def getTransitionById(self, transition_id):
        return self._id_transitions[transition_id]

class WorkflowState(object):
    implements(IWorkflowState)
    state_key = "hurry.workflow.state"
    id_key  = "hurry.workflow.id"

    def __init__(self, context):
        # XXX okay, I'm tired of it not being able to set annotations, so
        # we'll do this. Ugh.
        from zope.security.proxy import removeSecurityProxy
        self.context = removeSecurityProxy(context)
        self._annotations = IAnnotations(self.context)

    def initialize(self):
        wf_versions = component.queryUtility(IWorkflowVersions)
        if wf_versions is not None:
            self.setId(wf_versions.createVersionId())

    def setState(self, state):
        if state != self.getState():
            self._annotations[self.state_key] = state

    def setId(self, id):
        # XXX catalog should be informed (or should it?)
        self._annotations[self.id_key] = id

    def getState(self):
        return self._annotations.get(self.state_key, None)

    def getId(self):
        return self._annotations.get(self.id_key, None)

class WorkflowInfo(object):
    implements(IWorkflowInfo)
    name = u''

    def __init__(self, context):
        self.context = context
        self.wf = component.getUtility(IWorkflow, name=self.name)

    @classmethod
    def info(cls, obj):
        return component.getAdapter(obj, IWorkflowInfo, name=cls.name)

    @classmethod
    def state(cls, obj):
        return component.getAdapter(obj, IWorkflowState, name=cls.name)

    def fireTransition(self, transition_id, comment=None, side_effect=None,
                       check_security=True):
        state = self.state(self.context)
        # this raises InvalidTransitionError if id is invalid for current state
        transition = self.wf.getTransition(state.getState(), transition_id)
        # check whether we may execute this workflow transition
        try:
            interaction = getInteraction()
        except NoInteraction:
            checkPermission = nullCheckPermission
        else:
            if check_security:
                checkPermission = interaction.checkPermission
            else:
                checkPermission = nullCheckPermission
        if not checkPermission(
            transition.permission, self.context):
            raise Unauthorized(self.context,
                               'transition: %s' % transition_id,
                               transition.permission)
        # now make sure transition can still work in this context
        if not transition.condition(self, self.context):
            # XXX should we include state info here? if so, what?
            raise ConditionFailedError
        # perform action, return any result as new version
        result = transition.action(self, self.context)
        if result is not None:
            if transition.source is None:
                self.state(result).initialize()
            # stamp it with version
            state = self.state(result)
            state.setId(self.state(self.context).getId())
            # execute any side effect:
            if side_effect is not None:
                side_effect(result)
            event = WorkflowVersionTransitionEvent(result, self.context,
                                                   transition.source,
                                                   transition.destination,
                                                   transition, comment)
        else:
            if transition.source is None:
                self.state(self.context).initialize()
            # execute any side effect
            if side_effect is not None:
                side_effect(self.context)
            event = WorkflowTransitionEvent(self.context,
                                            transition.source,
                                            transition.destination,
                                            transition, comment)
        # change state of context or new object
        state.setState(transition.destination)
        notify(event)
        # send modified event for original or new object
        if result is None:
            notify(ObjectModifiedEvent(self.context))
        else:
            notify(ObjectModifiedEvent(result))
        return result

    def fireTransitionToward(self, state, comment=None, side_effect=None,
                             check_security=True):
        transition_ids = self.getFireableTransitionIdsToward(state,
                                                             check_security)
        if not transition_ids:
            raise interfaces.NoTransitionAvailableError(
                self.state(self.context).getState(),
                state)
        if len(transition_ids) != 1:
            raise interfaces.AmbiguousTransitionError(
                self.state(self.context).getState(),
                state)
        return self.fireTransition(transition_ids[0],
                                   comment, side_effect, check_security)

    def fireTransitionForVersions(self, state, transition_id):
        id = self.state(self.context).getId()
        wf_versions = component.getUtility(IWorkflowVersions)
        for version in wf_versions.getVersions(state, id):
            if version is self.context:
                continue
            self.info(version).fireTransition(transition_id)

    def fireAutomatic(self):
        for transition_id in self.getAutomaticTransitionIds():
            try:
                self.fireTransition(transition_id)
            except ConditionFailedError:
                # if condition failed, that's fine, then we weren't
                # ready to fire yet
                pass
            else:
                # if we actually managed to fire a transition,
                # we're done with this one now.
                return

    def hasVersion(self, state):
        wf_versions = component.getUtility(IWorkflowVersions)
        id = self.state(self.context).getId()
        return wf_versions.hasVersion(state, id)

    def getManualTransitionIds(self, check_security=True):
        try:
            checkPermission = getInteraction().checkPermission
        except NoInteraction:
            checkPermission = nullCheckPermission
        if not check_security:
            checkPermission = nullCheckPermission
        return [transition.transition_id for transition in
                sorted(self._getTransitions(MANUAL)) if
                transition.condition(self, self.context) and
                checkPermission(transition.permission, self.context)]

    def getSystemTransitionIds(self):
        # ignore permission checks
        return [transition.transition_id for transition in
                sorted(self._getTransitions(SYSTEM)) if
                transition.condition(self, self.context)]

    def getFireableTransitionIds(self, check_security=True):
        return (self.getManualTransitionIds(check_security) +
                self.getSystemTransitionIds())

    def getFireableTransitionIdsToward(self, state, check_security=True):
        result = []
        for transition_id in self.getFireableTransitionIds(check_security):
            transition = self.wf.getTransitionById(transition_id)
            if transition.destination == state:
                result.append(transition_id)
        return result

    def getAutomaticTransitionIds(self):
        return [transition.transition_id for transition in
                self._getTransitions(AUTOMATIC)]

    def hasAutomaticTransitions(self):
        # XXX could be faster
        return bool(self.getAutomaticTransitionIds())

    def _getTransitions(self, trigger):
        # retrieve all possible transitions from workflow utility
        transitions = self.wf.getTransitions(
            self.state(self.context).getState())
        # now filter these transitions to retrieve all possible
        # transitions in this context, and return their ids
        return [transition for transition in transitions if
                transition.trigger == trigger]

class WorkflowVersions(object):
    implements(IWorkflowVersions)

    def getVersions(self, state, id):
        raise NotImplementedError

    def getVersionsWithAutomaticTransitions(self):
        raise NotImplementedError

    def createVersionId(self):
        while True:
            id = random.randrange(sys.maxint)
            if not self.hasVersionId(id):
                return id
        assert False, "Shouldn't ever reach here"

    def hasVersion(self, state, id):
        raise NotImplementedError

    def hasVersionId(self, id):
        raise NotImplementedError

    def fireAutomatic(self):
        for version in self.getVersionsWithAutomaticTransitions():
            IWorkflowInfo(version).fireAutomatic()

class WorkflowTransitionEvent(ObjectEvent):
    implements(interfaces.IWorkflowTransitionEvent)

    def __init__(self, object, source, destination, transition, comment):
        super(WorkflowTransitionEvent, self).__init__(object)
        self.source = source
        self.destination = destination
        self.transition = transition
        self.comment = comment

class WorkflowVersionTransitionEvent(WorkflowTransitionEvent):
    implements(interfaces.IWorkflowVersionTransitionEvent)

    def __init__(self, object, old_object, source, destination,
                 transition, comment):
        super(WorkflowVersionTransitionEvent, self).__init__(
            object, source, destination, transition, comment)
        self.old_object = old_object



import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = (
    read('README.txt')
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    'Detailed Documentation\n'
    '**********************\n'
    + '\n' +
    read('src', 'hurry', 'workflow', 'workflow.txt')
    + '\n' +
    'Download\n'
    '********\n'
    )

setup(
    name="hurry.workflow",
    version='0.13.1',
    packages=find_packages('src'),
    
    package_dir= {'':'src'},
    
    namespace_packages=['hurry'],
    package_data = {
    '': ['*.txt', '*.zcml'],
    },

    zip_safe=False,
    author='Martijn Faassen (at Infrae)',
    author_email='faassen@startifact.com',
    description="""\
hurry.workflow is a simple workflow system. It can be used to
implement stateful multi-version workflows for Zope Toolkit applications.
""",
    long_description=long_description,
    license='ZPL 2.1',
    keywords="zope zope3",
    classifiers = ['Framework :: Zope3'],
    extras_require = dict(
        test = ['zope.testing'],
        ),
    install_requires=[
        'setuptools',
        'zope.interface',
        'zope.component',
        # this is an indirect dependency through zope.annotation, but
        # this at the time of working didn't yet declare this
        # dependency itself
        'ZODB3',
        'zope.event',
        'zope.security',
        'zope.annotation',
        'zope.lifecycleevent'],
    )



from zope.interface import Interface, Attribute
from zope.component.interfaces import IObjectEvent

MANUAL = 0
AUTOMATIC = 1
SYSTEM = 2

class InvalidTransitionError(Exception):
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return 'source: "%s"' % self.source
    
class NoTransitionAvailableError(InvalidTransitionError):
    def __init__(self, source, destination):
        super(NoTransitionAvailableError, self).__init__(source)
        self.destination = destination

    def __str__(self):
        return 'source: "%s" destination: "%s"' % (self.source, self.destination)
    
class AmbiguousTransitionError(InvalidTransitionError):
    def __init__(self, source, destination):
        super(AmbiguousTransitionError, self).__init__(source)
        self.destination = destination

    def __str__(self):
        return 'source: "%s" destination: "%s"' % (self.source, self.destination)

class ConditionFailedError(Exception):
    pass

class IWorkflow(Interface):
    """Defines workflow in the form of transition objects.

    Defined as a utility.
    """
    def initialize():
        """Do any needed initialization.

        Such as initialization with the workflow versions system.
        """

    def refresh(transitions):
        """Refresh workflow completely with new transitions.
        """
        
    def getTransitions(source):
        """Get all transitions from source.
        """

    def getTransition(source, transition_id):
        """Get transition with transition_id given source state.

        If the transition is invalid from this source state,
        an InvalidTransitionError is raised.
        """

    def getTransitionById(transition_id):
        """Get transition with transition_id.
        """

class IWorkflowState(Interface):
    """Store state on workflowed objects.

    Defined as an adapter.
    """
    
    def setState(state):
        """Set workflow state for this object.
        """
        
    def setId(id):
        """Set workflow version id for this object.

        This is used to mark all versions of an object with the
        same id.
        """
        
    def getState():
        """Return workflow state of this object.
        """

    def getId():
        """Get workflow version id for this object.

        This is used to mark all versions of an object with the same id.
        """
        
class IWorkflowInfo(Interface):
    """Get workflow info about workflowed object, and drive workflow.

    Defined as an adapter.
    """

    def setInitialState(state, comment=None):
        """Set initial state for the context object.

        Will also set a unique id for this new workflow.
        
        Fires a transition event.
        """
        
    def fireTransition(transition_id, comment=None, side_effect=None,
                       check_security=True):
        """Fire a transition for the context object.

        There's an optional comment parameter that contains some
        opaque object that offers a comment about the transition.
        This is useful for manual transitions where users can motivate
        their actions.

        There's also an optional side effect parameter which should
        be a callable which receives the object undergoing the transition
        as the parameter. This could do an editing action of the newly
        transitioned workflow object before an actual transition event is
        fired.

        If check_security is set to False, security is not checked
        and an application can fire a transition no matter what the
        user's permission is.
        """

    def fireTransitionToward(state, comment=None, side_effect=None,
                             check_security=True):
        """Fire transition toward state.

        Looks up a manual transition that will get to the indicated
        state.

        If no such transition is possible, NoTransitionAvailableError will
        be raised.

        If more than one manual transitions are possible,
        AmbiguousTransitionError will be raised.
        """
        
    def fireTransitionForVersions(state, transition_id):
        """Fire a transition for all versions in a state.
        """

    def fireAutomatic():
        """Fire automatic transitions if possible by condition.
        """
        
    def hasVersion(state):
        """Return true if a version exists in state.
        """
        
    def getManualTransitionIds():
        """Returns list of valid manual transitions.

        These transitions have to have a condition that's True.
        """

    def getManualTransitionIdsToward(state):
        """Returns list of manual transitions towards state.
        """
        
    def getAutomaticTransitionIds():
        """Returns list of possible automatic transitions.

        Condition is not checked.
        """
        
    def hasAutomaticTransitions():
        """Return true if there are possible automatic outgoing transitions.

        Condition is not checked.
        """

class IReadWorkflowVersions(Interface):
    
    def getVersions(state, id):
        """Get all versions of object known for this id and state.
        """

    def getVersionsWithAutomaticTransitions():
        """Get all versions that have outgoing transitions that are automatic.
        """

    def createVersionId():
        """Return new unique version id.
        """
        
    def hasVersion(id, state):
        """Return true if a version exists with the specific workflow state.
        """

    def hasVersionId(id):
        """Return True if version id is already in use.
        """

class IWriteWorkflowVersions(Interface):
    def fireAutomatic():
        """Fire all automatic transitions in the workflow (for all versions).
        """
    
class IWorkflowVersions(IReadWorkflowVersions, IWriteWorkflowVersions):
    """Interface to get information about versions of content in workflow.

    This can be implemented on top of the Zope catalog, for instance.

    Defined as a utility
    """

class IWorkflowTransitionEvent(IObjectEvent):
    source = Attribute('Original state or None if initial state')
    destination = Attribute('New state') 
    transition = Attribute('Transition that was fired or None if initial state')
    comment = Attribute('Comment that went with state transition')

class IWorkflowVersionTransitionEvent(IWorkflowTransitionEvent):
    old_object = Attribute('Old version of object')
